# this file is used by basically everything
# so if it's in literally any other place, everything dies from circular deps
# basically, just leave it here

from __future__ import annotations

import json
import logging
import re
from fnmatch import fnmatch
from pathlib import Path
from typing import Annotated, Any, ClassVar, Literal, Self, TypeVar

from nbtlib import (
    Compound,
    Path as NBTPath,
    parse_nbt,  # pyright: ignore[reportUnknownVariableType]
)
from pydantic import (
    BeforeValidator,
    ConfigDict,
    JsonValue,
    TypeAdapter,
    field_validator,
    model_serializer,
    model_validator,
)
from pydantic.config import JsonDict
from pydantic.dataclasses import dataclass
from pydantic.functional_validators import ModelWrapValidatorHandler
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import override

from hexdoc.model import DEFAULT_CONFIG
from hexdoc.utils import TRACE

logger = logging.getLogger(__name__)

ResourceType = Literal["assets", "data", ""]

_T = TypeVar("_T")

MODEL_PATH_REGEX = re.compile(
    r"""
    assets
    /(?P<namespace>[a-z0-9_\-.]+)
    /(?:
        models/(?P<model_type>[a-z0-9_\-.]+)
        | blockstates
    )
    /(?P<path>[a-z0-9_\-./]+)\.json""",
    re.VERBOSE,
)


def _make_regex(count: bool = False, nbt: bool = False) -> re.Pattern[str]:
    pattern = r"(?:(?P<namespace>[0-9a-z_\-.]+):)?(?P<path>[0-9a-z_\-./*]+)"
    if count:
        pattern += r"(?:#(?P<count>[0-9]+))?"
    if nbt:
        pattern += r"(?P<nbt>{.*})?"
    return re.compile(pattern)


def resloc_json_schema_extra(
    schema: JsonDict,
    model_type: type[BaseResourceLocation],
):
    object_schema = schema.copy()

    regex = model_type._from_str_regex  # pyright: ignore[reportPrivateUsage]
    pattern = re.sub(
        r"\(\?P<(.+?)>(.+?)\)",
        r"(?<\1>\2)",
        f"^{regex.pattern}$",
    )
    string_schema: JsonDict = {
        "type": "string",
        "pattern": pattern,
    }

    schema.clear()

    for key in ["title", "description"]:
        if value := object_schema.pop(key, None):
            schema[key] = value

    schema.update(string_schema)


@dataclass(
    frozen=True,
    repr=False,
    config=DEFAULT_CONFIG
    | ConfigDict(
        json_schema_extra=resloc_json_schema_extra,
        arbitrary_types_allowed=True,
    ),
)
class BaseResourceLocation:
    namespace: str
    path: str

    _from_str_regex: ClassVar[re.Pattern[str]]

    def __init_subclass__(cls, regex: re.Pattern[str] | None) -> None:
        if regex:
            cls._from_str_regex = regex

    @classmethod
    def from_str(cls, raw: str) -> Self:
        match = cls._from_str_regex.fullmatch(raw)
        if match is None:
            raise ValueError(f"Invalid {cls.__name__} string: {raw}")

        return cls(**match.groupdict())

    @classmethod
    def model_validate(cls, value: Any, *, context: Any = None):
        ta = TypeAdapter(cls)
        return ta.validate_python(value, context=context)

    @model_validator(mode="wrap")
    @classmethod
    def _pre_root(cls, values: Any, handler: ModelWrapValidatorHandler[Self]):
        # before validating the fields, if it's a string instead of a dict, convert it
        logger.log(TRACE, f"Convert {values} to {cls.__name__}")
        if isinstance(values, str):
            return cls.from_str(values)
        return handler(values)

    @field_validator("namespace", mode="before")
    def _default_namespace(cls, value: Any):
        match value:
            case str():
                return value.lower()
            case None:
                return "minecraft"
            case _:
                return value

    @field_validator("path")
    def _validate_path(cls, value: str):
        return value.lower().rstrip("/")

    @model_serializer
    def _ser_model(self) -> str:
        return str(self)

    @property
    def id(self) -> ResourceLocation:
        return ResourceLocation(self.namespace, self.path)

    def i18n_key(self, root: str) -> str:
        # TODO: is this how i18n works????? (apparently, because it's working)
        return f"{root}.{self.namespace}.{self.path.replace('/', '.')}"

    def __repr__(self) -> str:
        return f"{self.namespace}:{self.path}"


@dataclass(frozen=True, repr=False)
class ResourceLocation(BaseResourceLocation, regex=_make_regex()):
    """Represents a Minecraft resource location / namespaced ID."""

    is_tag: bool = False

    @classmethod
    def from_str(cls, raw: str) -> Self:
        id = super().from_str(raw.removeprefix("#"))
        if raw.startswith("#"):
            object.__setattr__(id, "is_tag", True)
        return id

    @classmethod
    def from_file(cls, modid: str, base_dir: Path, path: Path) -> Self:
        resource_path = path.relative_to(base_dir).with_suffix("").as_posix()
        return cls(modid, resource_path)

    @classmethod
    def from_model_path(cls, model_path: str | Path) -> Self:
        match = MODEL_PATH_REGEX.search(Path(model_path).as_posix())
        if not match:
            raise ValueError(f"Failed to match model path: {model_path}")
        return cls(match["namespace"], match["path"])

    @property
    def href(self) -> str:
        return f"#{self.path}"

    @property
    def css_class(self) -> str:
        stripped_path = re.sub(r"[\*\/\.]", "-", self.path)
        return f"texture-{self.namespace}-{stripped_path}"

    def with_namespace(self, namespace: str) -> Self:
        """Returns a copy of this ResourceLocation with the given namespace."""
        return self.__class__(namespace, self.path)

    def with_path(self, path: str | Path) -> Self:
        """Returns a copy of this ResourceLocation with the given path."""
        if isinstance(path, Path):
            path = path.as_posix()
        return self.__class__(self.namespace, path)

    def match(self, pattern: Self) -> bool:
        return fnmatch(str(self), str(pattern))

    def template_path(self, type: str, folder: str = "") -> str:
        return self.file_path_stub(type, folder, assume_json=False).as_posix()

    def file_path_stub(
        self,
        type: ResourceType | str,
        folder: str | Path = "",
        assume_json: bool = True,
    ) -> Path:
        """Returns the path to find this resource within a resource directory.

        If `assume_json` is True and no file extension is provided, `.json` is assumed.

        For example:
        ```py
        ResLoc("hexcasting", "thehexbook/book").file_path_stub("data", "patchouli_books")
        # data/hexcasting/patchouli_books/thehexbook/book.json
        ```
        """
        # if folder is an empty string, Path won't add an extra slash
        path = Path(type) / self.namespace / folder / self.path
        if assume_json and not path.suffix:
            return path.with_suffix(".json")
        return path

    def removeprefix(self, prefix: str) -> Self:
        return self.with_path(self.path.removeprefix(prefix))

    def __truediv__(self, other: str) -> Self:
        return self.with_path(f"{self.path}/{other}")

    def __rtruediv__(self, other: str) -> Self:
        return self.with_path(f"{other}/{self.path}")

    def __add__(self, other: str) -> Self:
        return self.with_path(self.path + other)

    def __repr__(self) -> str:
        s = super().__repr__()
        if self.is_tag:
            return f"#{s}"
        return s


# pure unadulterated laziness
ResLoc = ResourceLocation


# FIXME: commas?????????????????
# https://vazkiimods.github.io/Patchouli/docs/patchouli-advanced/itemstack-format
@dataclass(frozen=True, repr=False)
class ItemStack(BaseResourceLocation, regex=_make_regex(count=True, nbt=True)):
    """Represents an item with optional count and NBT.

    Inherits from BaseResourceLocation, not ResourceLocation.
    """

    count: int | None = None
    nbt: str | None = None

    _data: SkipJsonSchema[Compound | None] = None

    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(regex=cls._from_str_regex, **kwargs)

    def __post_init__(self):
        object.__setattr__(self, "_data", _parse_nbt(self.nbt))

    @property
    def data(self):
        return self._data

    def get_name(self) -> str | None:
        if self.data is None:
            return None

        component_json = self.data.get(NBTPath("display.Name"))  # pyright: ignore[reportUnknownVariableType, reportUnknownMemberType]
        if not isinstance(component_json, str):
            return None

        try:
            component: JsonValue = json.loads(component_json)
        except ValueError:
            return None

        if not isinstance(component, dict):
            return None

        name = component.get("text")
        if not isinstance(name, str):
            return None

        return name

    @override
    def i18n_key(self, root: str = "item") -> str:
        return super().i18n_key(root)

    def __repr__(self) -> str:
        s = super().__repr__()
        if self.count is not None:
            s += f"#{self.count}"
        if self.nbt is not None:
            s += self.nbt
        return s


@dataclass(frozen=True, repr=False)
class Entity(BaseResourceLocation, regex=_make_regex(nbt=True)):
    """Represents an entity with optional NBT.

    Inherits from BaseResourceLocation, not ResourceLocation.
    """

    nbt: str | None = None

    def __repr__(self) -> str:
        s = super().__repr__()
        if self.nbt is not None:
            s += self.nbt
        return s


def _add_hashtag_to_tag(value: Any):
    match value:
        case str() if not value.startswith("#"):
            return f"#{value}"
        case ResourceLocation(namespace, path) if not value.is_tag:
            return ResourceLocation(namespace, path, is_tag=True)
        case _:
            return value


AssumeTag = Annotated[_T, BeforeValidator(_add_hashtag_to_tag)]
"""Validator that adds `#` to the start of strings, and sets `ResourceLocation.is_tag`
to `True`."""


def _parse_nbt(nbt: str | None) -> Compound | None:
    if nbt is None:
        return None

    # TODO: maybe re-add strict parsing when we're sure it will actually work?
    try:
        result = parse_nbt(nbt)
    except Exception as e:
        logger.warning(f"Failed to parse sNBT literal '{nbt}': {e}")
        return None

    if not isinstance(result, Compound):
        logger.warning(f"Expected Compound, got {type(result)}: {result}")
        return None

    return result



# pyright: reportPrivateUsage=false

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import defaultdict
from typing import Any, ClassVar, Generator, Self, Unpack

import more_itertools
from pydantic import (
    ConfigDict,
    GetJsonSchemaHandler,
    TypeAdapter,
    ValidationError,
    ValidationInfo,
    model_validator,
)
from pydantic.functional_validators import ModelWrapValidatorHandler
from pydantic.json_schema import JsonSchemaValue
from pydantic_core import InitErrorDetails, PydanticCustomError, core_schema as cs
from typing_extensions import override

from hexdoc.core.resource import ResourceLocation
from hexdoc.plugin import PluginManager
from hexdoc.utils import (
    Inherit,
    InheritType,
    NoValue,
    NoValueType,
    classproperty,
)

from .base import HexdocModel

TagValue = str | NoValueType

_RESOLVED = "__resolved"

# sentinel value to check if we already loaded the tagged union subtypes hook
_is_loaded = False


class InternallyTaggedUnion(HexdocModel):
    """Implements [internally tagged unions](https://serde.rs/enum-representations.html#internally-tagged)
    using the [Registry pattern](https://charlesreid1.github.io/python-patterns-the-registry.html).

    To ensure your subtypes are loaded even if they're not imported by any file, add a
    Pluggy hook implementation for `hexdoc_load_tagged_unions() -> list[Package]`.

    Subclasses MUST NOT be generic unless they provide a default value for all
    `__init_subclass__` arguments. See pydantic/7171 for more info.

    Args:
        key: The dict key for the internal tag. If None, the parent's value is used.
        value: The expected tag value for this class. Should be None for types which
            shouldn't be instantiated (eg. abstract classes).
    """

    # inherited
    _tag_key: ClassVar[str | None] = None
    _tag_value: ClassVar[TagValue | None] = None

    # per-class
    __all_subtypes: ClassVar[set[type[Self]]]
    __concrete_subtypes: ClassVar[defaultdict[TagValue, set[type[Self]]]]

    def __init_subclass__(
        cls,
        *,
        key: str | InheritType | None = Inherit,
        value: TagValue | InheritType | None = Inherit,
        **kwargs: Unpack[ConfigDict],
    ):
        super().__init_subclass__(**kwargs)

        # inherited data
        if key is not Inherit:
            cls._tag_key = key
        if value is not Inherit:
            cls._tag_value = value

        # don't bother with rest of init if it's not part of a union
        if cls._tag_key is None:
            if cls._tag_value is None:
                return
            raise ValueError(
                f"Expected value=None for {cls} with key=None, got {value}"
            )

        # per-class data and lookups
        cls.__all_subtypes = set()
        cls.__concrete_subtypes = defaultdict(set)

        # add to all the parents
        for supertype in cls._supertypes():
            supertype.__all_subtypes.add(cls)
            if cls._tag_value is not None:
                supertype.__concrete_subtypes[cls._tag_value].add(cls)

    @classmethod
    def _tag_key_or_raise(cls) -> str:
        if cls._tag_key is None:
            raise NotImplementedError
        return cls._tag_key

    @classmethod
    def _supertypes(cls) -> Generator[type[InternallyTaggedUnion], None, None]:
        tag_key = cls._tag_key_or_raise()

        # we consider a type to be its own supertype/subtype
        yield cls

        # recursively yield bases
        # stop when we reach a non-union or a type with a different key (or no key)
        for base in cls.__bases__:
            if issubclass(base, InternallyTaggedUnion) and base._tag_key == tag_key:
                yield from base._supertypes()

    @model_validator(mode="wrap")
    @classmethod
    def _resolve_from_dict(
        cls,
        value: Any,
        handler: ModelWrapValidatorHandler[Self],
        info: ValidationInfo,
    ) -> Self:
        pm = PluginManager.of(info)

        # load plugins from entry points
        global _is_loaded
        if not _is_loaded:
            more_itertools.consume(pm.load_tagged_unions())
            _is_loaded = True

        # do this early so we know it's part of a union before returning anything
        tag_key = cls._tag_key_or_raise()

        # if it's already instantiated, just return it; otherwise ensure it's a dict
        match value:
            case InternallyTaggedUnion() if isinstance(value, cls):
                return value
            case dict() if _RESOLVED not in value:
                data: dict[str, Any] = value
                data[_RESOLVED] = True
            case _:
                return handler(value)

        # tag value, eg. "minecraft:crafting_shaped"
        tag_value = data.get(tag_key, NoValue)

        # list of matching types, eg. [ShapedCraftingRecipe, ModConditionalShapedCraftingRecipe]
        tag_types = cls.__concrete_subtypes.get(tag_value)
        if tag_types is None:
            raise TypeError(f"Unhandled tag: {tag_key}={tag_value} for {cls}: {data}")

        # try all the types
        exceptions: list[InitErrorDetails] = []
        matches: dict[type[Self], Self] = {}

        for inner_type in tag_types:
            try:
                matches[inner_type] = inner_type.model_validate(
                    data, context=info.context
                )
            except Exception as e:
                exceptions.append(
                    InitErrorDetails(
                        type=PydanticCustomError(
                            "TaggedUnionMatchError",
                            "{exception_class}: {exception}",
                            {
                                "exception_class": e.__class__.__name__,
                                "exception": str(e),
                            },
                        ),
                        loc=(
                            cls.__name__,
                            inner_type.__name__,
                        ),
                        input=data,
                    )
                )

        # ensure we only matched one
        match len(matches):
            case 1:
                return matches.popitem()[1]
            case x if x > 1:
                ambiguous_types = ", ".join(str(t) for t in matches.keys())
                reason = f"Ambiguous union match: {ambiguous_types}"
            case _:
                reason = "No match found"

        # something went wrong, raise an exception
        error = PydanticCustomError(
            "TaggedUnionMatchErrorGroup",
            (
                "Failed to match tagged union {class_name}: {reason}\n"
                "  Tag: {tag_key}={tag_value}\n"
                "  Types: {types}\n"
                "  Data: {data}"
            ),
            {
                "class_name": str(cls),
                "reason": reason,
                "tag_key": cls._tag_key,
                "tag_value": tag_value,
                "types": ", ".join(str(t) for t in tag_types),
                "data": repr(data),
            },
        )

        if exceptions:
            if _RESOLVED in data:
                data.pop(_RESOLVED)  # avoid interfering with other types
            exceptions.insert(
                0,
                InitErrorDetails(
                    type=error,
                    loc=(cls.__name__,),
                    input=data,
                ),
            )
            raise ValidationError.from_exception_data(
                "TaggedUnionMatchError", exceptions
            )

        raise RuntimeError(str(error))

    @model_validator(mode="before")
    def _pop_temporary_keys(cls, value: dict[Any, Any] | Any):
        if isinstance(value, dict) and _RESOLVED in value:
            # copy because this validator may be called multiple times
            # eg. two types with the same key
            value = value.copy()
            value.pop(_RESOLVED)
            assert value.pop(cls._tag_key, NoValue) == cls._tag_value
        return value

    @classmethod
    @override
    def __get_pydantic_json_schema__(
        cls,
        core_schema: cs.CoreSchema,
        handler: GetJsonSchemaHandler,
    ) -> JsonSchemaValue:
        base_schema = handler.resolve_ref_schema(
            super().__get_pydantic_json_schema__(core_schema, handler)
        )

        if cls._tag_key is None:
            return base_schema

        properties = base_schema.setdefault("properties", {})

        properties[cls._tag_key] = handler(cs.literal_schema([cls._tag_value]))
        base_schema.setdefault("required", []).append(cls._tag_key)

        if cls._tag_value is not None:
            return base_schema

        subtypes = {
            subtype
            for subtypes in cls.__concrete_subtypes.values()
            for subtype in subtypes
            if subtype is not cls
        }
        if not subtypes:
            return base_schema

        union_schema = cs.union_schema(
            [subtype.__pydantic_core_schema__ for subtype in subtypes],
        )
        json_schema = handler.resolve_ref_schema(handler(union_schema))

        if any_of := json_schema.get("anyOf"):
            other_schema = base_schema | {
                "additionalProperties": True,
                "properties": (
                    properties
                    | {
                        cls._tag_key: {
                            "allOf": [
                                handler(cls._tag_value_schema),
                                {
                                    "not": handler(
                                        cs.union_schema(
                                            [
                                                cs.literal_schema([subtype._tag_value])
                                                for subtype in subtypes
                                                if subtype._tag_value
                                                not in {None, NoValue}
                                            ]
                                        )
                                    )
                                },
                            ],
                        },
                    }
                ),
            }
            any_of.append(other_schema)

        return json_schema

    @classproperty
    @classmethod
    def _tag_value_type(cls) -> type[Any]:
        return str

    @classproperty
    @classmethod
    def _tag_value_schema(cls):
        return TypeAdapter(cls._tag_value_type).core_schema


class TypeTaggedUnion(InternallyTaggedUnion, key="type", value=None):
    _type: ClassVar[ResourceLocation | NoValueType | None] = None

    def __init_subclass__(
        cls,
        *,
        type: TagValue | InheritType | None = Inherit,
        **kwargs: Unpack[ConfigDict],
    ):
        super().__init_subclass__(value=type, **kwargs)

        match cls._tag_value:
            case str(raw_value):
                cls._type = ResourceLocation.from_str(raw_value)
            case value:
                cls._type = value

    @classproperty
    @classmethod
    def type(cls):
        return cls._type

    @classproperty
    @classmethod
    @override
    def _tag_value_type(cls) -> type[Any]:
        return ResourceLocation


class TypeTaggedTemplate(TypeTaggedUnion, ABC, type=None):
    __template_id: ClassVar[ResourceLocation]

    def __init_subclass__(
        cls,
        *,
        type: str | InheritType | None = Inherit,
        template_type: str | None = None,
        **kwargs: Unpack[ConfigDict],
    ) -> None:
        super().__init_subclass__(type=type, **kwargs)

        # jinja template path
        if template_type is not None:
            template_id = ResourceLocation.from_str(template_type)
        else:
            template_id = cls.type

        if template_id:
            cls.__template_id = template_id

    @classproperty
    @classmethod
    @abstractmethod
    def template(cls) -> str:
        """Returns the Jinja template path for this class without any file extension.

        For example, return `"pages/{path}"`, not `"pages/{path}.html.jinja"`.
        """

    @classproperty
    @classmethod
    def template_id(cls):
        return cls.__template_id



from hexdoc.core import ResourceLocation
from hexdoc.minecraft import LocalizedStr
from hexdoc.minecraft.assets import ItemWithTexture, NamedTexture
from hexdoc.model import HexdocModel
from pydantic import Field

from .condition import Condition
from .page import Page


class EntryParent(HexdocModel):
    """https://klikli-dev.github.io/modonomicon/docs/basics/structure/entries#parents"""

    entry: ResourceLocation
    draw_arrow: bool = True
    line_enabled: bool = True
    line_reversed: bool = False


class Entry(HexdocModel):
    """https://klikli-dev.github.io/modonomicon/docs/basics/structure/entries"""

    category: ResourceLocation
    name: LocalizedStr
    description: LocalizedStr
    icon: ItemWithTexture | NamedTexture
    x: int
    y: int

    hide_while_locked: bool = False
    background_u_index: int = 0
    background_v_index: int = 0
    condition: Condition | None = None
    parents: list[EntryParent] = Field(default_factory=list)
    pages: list[Page] = Field(default_factory=list)
    category_to_open: ResourceLocation | None = None
    command_to_run_on_first_read: ResourceLocation | None = None



import functools
from typing import Callable, Iterator, ParamSpec, TypeVar

_T = TypeVar("_T")
_P = ParamSpec("_P")


def must_yield_something(f: Callable[_P, Iterator[_T]]) -> Callable[_P, Iterator[_T]]:
    """Raises StopIteration if the wrapped iterator doesn't yield anything."""

    @functools.wraps(f)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> Iterator[_T]:
        iterator = f(*args, **kwargs)
        yield next(iterator)
        yield from iterator

    return wrapper


def listify(f: Callable[_P, Iterator[_T]]) -> Callable[_P, list[_T]]:
    @functools.wraps(f)
    def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> list[_T]:
        return list(f(*args, **kwargs))

    return wrapper



from collections import defaultdict
from typing import Self

from jinja2 import pass_context
from jinja2.runtime import Context
from pydantic import (
    Field,
    PrivateAttr,
    ValidationInfo,
    field_validator,
    model_validator,
)
from typing_extensions import override

from hexdoc.core import Entity, ItemStack, ResourceLocation
from hexdoc.minecraft import I18n, LocalizedStr
from hexdoc.minecraft.assets import ItemWithTexture, PNGTexture, TagWithTexture, Texture
from hexdoc.minecraft.assets.items import ImageTexture
from hexdoc.minecraft.assets.load_assets import validate_texture
from hexdoc.minecraft.recipe import (
    BlastingRecipe,
    CampfireCookingRecipe,
    CraftingRecipe,
    SmeltingRecipe,
    SmithingRecipe,
    SmokingRecipe,
    StonecuttingRecipe,
)
from hexdoc.model import HexdocModel

from ..text import FormatTree
from .abstract_pages import (
    Page,
    PageWithDoubleRecipe,
    PageWithDoubleRecipeAccumulator,
    PageWithText,
    PageWithTitle,
    RecipeAccumulatorPage,
)


class TextPage(Page, type="patchouli:text"):
    title: LocalizedStr | None = None
    text: FormatTree


class BlastingPage(PageWithDoubleRecipe[BlastingRecipe], type="patchouli:blasting"):
    pass


class CampfireCookingPage(
    PageWithDoubleRecipe[CampfireCookingRecipe], type="patchouli:campfire_cooking"
):
    pass


class CraftingPage(
    PageWithDoubleRecipeAccumulator[CraftingRecipe], type="patchouli:crafting"
):
    @classmethod
    @override
    def accumulator_type(cls):
        return CraftingAccumulatorPage


class CraftingAccumulatorPage(
    RecipeAccumulatorPage[CraftingPage, CraftingRecipe], page_type=CraftingPage
):
    pass


class EmptyPage(Page, type="patchouli:empty", template_type="patchouli:page"):
    draw_filler: bool = True


class EntityPage(PageWithText, type="patchouli:entity"):
    _entity_name: LocalizedStr = PrivateAttr()
    _texture: PNGTexture = PrivateAttr()

    entity: Entity
    scale: float = 1
    offset: float = 0
    rotate: bool = True
    default_rotation: float = -45
    name_field: LocalizedStr | None = Field(default=None, serialization_alias="name")

    @property
    def entity_name(self):
        return self._entity_name

    @property
    def name(self):
        if self.name_field is None or not self.name_field.value:
            return self._entity_name
        return self.name_field

    @property
    def texture(self):
        return self._texture

    @model_validator(mode="after")
    def _get_texture(self, info: ValidationInfo) -> Self:
        # can't be on Entity's validator because it's frozen and
        # causes circular references with the PNGTexture
        assert info.context is not None
        i18n = I18n.of(info)
        self._entity_name = i18n.localize_entity(self.entity.id)
        self._texture = PNGTexture.load_id(
            id="textures/entities" / self.entity.id + ".png", context=info.context
        )
        return self


class ImagePage(PageWithTitle, type="patchouli:image"):
    images: list[Texture]
    border: bool = False

    @property
    def images_with_alt(self):
        for image in self.images:
            if self.title:
                yield image, self.title
            else:
                yield image, str(image)


class LinkPage(TextPage, type="patchouli:link"):
    url: str
    link_text: LocalizedStr


class Multiblock(HexdocModel):
    """https://vazkiimods.github.io/Patchouli/docs/patchouli-basics/multiblocks/"""

    mapping: dict[str, ItemWithTexture | TagWithTexture]
    pattern: list[list[str]]
    symmetrical: bool = False
    offset: tuple[int, int, int] | None = None

    def bill_of_materials(self):
        character_counts = defaultdict[str, int](int)

        for layer in self.pattern:
            for row in layer:
                for character in row:
                    match character:
                        case str() if character in self.mapping:
                            character_counts[character] += 1
                        case " " | "0":  # air
                            pass
                        case _:
                            raise ValueError(
                                f"Character not found in multiblock mapping: `{character}`"
                            )

        materials = [
            (self.mapping[character], count)
            for character, count in character_counts.items()
        ]

        # sort by descending count, break ties by ascending name
        materials.sort(key=lambda v: v[0].name.value)
        materials.sort(key=lambda v: v[1], reverse=True)

        return materials

    @field_validator("mapping", mode="after")
    @classmethod
    def _add_default_mapping(
        cls,
        mapping: dict[str, ItemWithTexture | TagWithTexture],
        info: ValidationInfo,
    ):
        i18n = I18n.of(info)
        return {
            "_": ItemWithTexture(
                id=ItemStack("hexdoc", "any"),
                name=i18n.localize("hexdoc.any_block"),
                texture=PNGTexture.load_id(
                    ResourceLocation("hexdoc", "textures/gui/any_block.png"),
                    context=info,
                ),
            ),
        } | mapping


class MultiblockPage(PageWithText, type="patchouli:multiblock"):
    name: LocalizedStr
    multiblock_id: ResourceLocation | None = None
    multiblock: Multiblock | None = None
    enable_visualize: bool = True

    _texture: ImageTexture | None = PrivateAttr(None)

    @property
    def texture(self):
        return self._texture

    @property
    def _texture_id(self):
        if self.multiblock_id:
            return self.multiblock_id.with_path(
                f"textures/multiblock/hexdoc/{self.multiblock_id.path}.png"
            )

    @model_validator(mode="after")
    def _check_multiblock(self, info: ValidationInfo) -> Self:
        if self.multiblock_id is None and self.multiblock is None:
            raise ValueError(f"One of multiblock_id or multiblock must be set\n{self}")

        if texture_id := self._texture_id:
            self._texture = validate_texture(
                texture_id,
                context=info,
                model_type=ImageTexture,
            )

        return self


class QuestPage(PageWithText, type="patchouli:quest"):
    trigger: ResourceLocation | None = None
    title: LocalizedStr = LocalizedStr.with_value("Objective")


class RelationsPage(PageWithText, type="patchouli:relations"):
    entries: list[ResourceLocation]
    title: LocalizedStr = LocalizedStr.with_value("Related Chapters")

    @pass_context
    def get_entries(self, context: Context) -> list[ResourceLocation]:
        for entry in self.entries:
            if entry not in context["book"].all_entries:
                raise ValueError(f"Broken entry reference in relations: {entry}")
        return self.entries


class SmeltingPage(PageWithDoubleRecipe[SmeltingRecipe], type="patchouli:smelting"):
    pass


class SmithingPage(PageWithDoubleRecipe[SmithingRecipe], type="patchouli:smithing"):
    pass


class SmokingPage(PageWithDoubleRecipe[SmokingRecipe], type="patchouli:smoking"):
    pass


class StonecuttingPage(
    PageWithDoubleRecipe[StonecuttingRecipe], type="patchouli:stonecutting"
):
    pass


class SpotlightPage(PageWithText, type="patchouli:spotlight"):
    title_field: LocalizedStr | None = Field(default=None, alias="title")
    item: ItemWithTexture
    link_recipe: bool = False

    @property
    def title(self) -> LocalizedStr | None:
        return self.title_field or self.item.name



from functools import cached_property
from typing import Any, Literal, Self

from pydantic import Field

from hexdoc.model import HexdocModel
from hexdoc.utils.types import PydanticURL

from .textures import BaseTexture


class AnimationMetaFrame(HexdocModel):
    index: int | None = None
    time: int | None = None


class AnimationMetaTag(HexdocModel):
    interpolate: Literal[False] = False  # TODO: handle interpolation
    width: None = None  # TODO: handle non-square textures
    height: None = None
    frametime: int = 1
    frames: list[int | AnimationMetaFrame] = Field(default_factory=list)


class AnimationMeta(HexdocModel):
    animation: AnimationMetaTag


class AnimatedTextureFrame(HexdocModel):
    index: int
    start: int
    time: int
    animation_time: int

    @property
    def start_percent(self):
        return self._format_time(self.start)

    @property
    def end_percent(self):
        return self._format_time(self.start + self.time, backoff=True)

    def _format_time(self, time: int, *, backoff: bool = False) -> str:
        percent = 100 * time / self.animation_time
        if backoff and percent < 100:
            percent -= 0.01
        return f"{percent:.2f}".rstrip("0").rstrip(".")


class AnimatedTexture(BaseTexture):
    url: PydanticURL | None
    pixelated: bool
    css_class: str
    meta: AnimationMeta

    @classmethod
    def from_url(cls, *args: Any, **kwargs: Any) -> Self:
        raise NotImplementedError("AnimatedTexture does not support from_url()")

    @property
    def time_seconds(self):
        return self.time / 20

    @cached_property
    def time(self):
        return sum(time for _, time in self._normalized_frames)

    @property
    def frames(self):
        start = 0
        for index, time in self._normalized_frames:
            yield AnimatedTextureFrame(
                index=index,
                start=start,
                time=time,
                animation_time=self.time,
            )
            start += time

    @property
    def _normalized_frames(self):
        """index, time"""
        animation = self.meta.animation

        for i, frame in enumerate(animation.frames):
            match frame:
                case int(index):
                    time = None
                case AnimationMetaFrame(index=index, time=time):
                    pass

            if index is None:
                index = i
            if time is None:
                time = animation.frametime

            yield index, time



from abc import ABC
from typing import Annotated, Any, Iterator

from pydantic import (
    AfterValidator,
    BeforeValidator,
    Field,
    ValidationError,
    ValidationInfo,
)

from hexdoc.core import ResourceLocation
from hexdoc.core.loader import ModResourceLoader
from hexdoc.core.resource import AssumeTag
from hexdoc.model import HexdocModel, NoValue, TypeTaggedUnion
from hexdoc.utils import listify

from ..assets import ItemWithTexture, TagWithTexture
from ..tags import Tag


class ItemIngredient(TypeTaggedUnion, ABC):
    @property
    def item(self) -> ItemWithTexture | TagWithTexture: ...


class MinecraftItemIdIngredient(ItemIngredient, type=NoValue):
    item_: ItemWithTexture = Field(alias="item")

    @property
    def item(self):
        return self.item_


class MinecraftItemTagIngredient(ItemIngredient, type=NoValue):
    tag: AssumeTag[TagWithTexture]

    @property
    def item(self):
        return self.tag


class ItemResult(HexdocModel):
    item: ItemWithTexture
    count: int = 1


def _validate_single_item_to_list(value: Any):
    match value:
        case [*contents]:
            return contents
        case _:
            return [value]


@listify
def _validate_flatten_nested_tags(
    ingredients: list[ItemIngredient],
    info: ValidationInfo,
) -> Iterator[ItemIngredient]:
    loader = ModResourceLoader.of(info)
    for ingredient in ingredients:
        yield ingredient

        if isinstance(ingredient, MinecraftItemTagIngredient):
            yield from _items_in_tag(ingredient.tag.id, info, loader)


def _items_in_tag(
    tag_id: ResourceLocation,
    info: ValidationInfo,
    loader: ModResourceLoader,
) -> Iterator[ItemIngredient]:
    try:
        tag = Tag.load("items", tag_id, loader)
    except FileNotFoundError:
        return

    for id in tag.value_ids:
        try:
            yield MinecraftItemIdIngredient.model_validate(
                {"item": id},
                context=info.context,
            )
        except ValidationError:
            yield from _items_in_tag(id, info, loader)


ItemIngredientList = Annotated[
    list[ItemIngredient],
    BeforeValidator(_validate_single_item_to_list),
    AfterValidator(_validate_flatten_nested_tags),
]
"""A list of ItemIngredients. Accepts a single value or a list."""



from typing import Any, Callable, Iterable

from pydantic import ConfigDict, JsonValue


def inherited(
    schema: dict[str, Any],
    inherit: Iterable[str] = ("title", "description"),
):
    return {key: schema[key] for key in inherit if key in schema}


def type_str(schema: dict[str, Any]):
    return {
        "type": "string",
    }


def json_schema_extra(
    *fns: Callable[[dict[str, Any]], dict[str, Any]],
    replace: bool = False,
    **extra: JsonValue,
):
    def inner(schema: dict[str, Any]):
        new_schema = dict[str, Any]()
        for fn in fns:
            new_schema |= fn(schema)
        new_schema |= extra

        if replace:
            schema.clear()

        schema.update(new_schema)

    return inner


def json_schema_extra_config(
    *fns: Callable[[dict[str, Any]], dict[str, Any]],
    replace: bool = True,
    **extra: JsonValue,
) -> ConfigDict:
    return {"json_schema_extra": json_schema_extra(*fns, replace=replace, **extra)}



from typing import Any, Self, cast, overload

from jinja2.runtime import Context
from pydantic import ValidationInfo

from .classproperties import classproperty
from .deserialize import cast_or_raise

ContextSource = dict[str, Any] | ValidationInfo | Context
"""Valid argument types for `ValidationContext.of`."""


class ValidationContext:
    @classproperty
    @classmethod
    def context_key(cls) -> str:
        return str(cls)

    @overload
    @classmethod
    def of(cls, info: ValidationInfo, /) -> Self: ...

    @overload
    @classmethod
    def of(cls, context: dict[str, Any] | Context, /) -> Self: ...

    @overload
    @classmethod
    def of(cls, source: ContextSource, /) -> Self: ...

    @classmethod
    def of(cls, source: ContextSource, /) -> Self:
        match source:
            case dict() | Context():
                pass
            case _:
                source = cast_or_raise(source.context, dict)
        return cast_or_raise(source[cls.context_key], cls)

    def add_to_context(self, context: dict[str, Any], overwrite: bool = False):
        return add_to_context(context, self.context_key, self, overwrite)


def add_to_context(
    context: dict[str, Any],
    key: str,
    value: Any,
    overwrite: bool = False,
):
    if not overwrite and key in context:
        raise KeyError(f"Key {key} for {value} already exists in context")
    context[key] = value


def cast_context(source: ContextSource) -> dict[str, Any]:
    """Wrapper for `typing.cast` to simplify passing `ContextSource` to validation
    methods. This is a lie to keep the type checker happy."""
    return cast(dict[str, Any], source)



"""Custom data model for favicon properties."""

# Standard Library
import json as _json
from typing import Optional, Tuple, TypedDict


class FaviconDict(TypedDict):
    image_format: str
    dimensions: tuple[int, int]
    prefix: str
    rel: str | None


class FaviconProperties:
    """Data Model for Favicon Properties."""

    def __init__(
        self,
        image_fmt: str,
        dimensions: Tuple[int, int],
        prefix: str,
        rel: Optional[str] = None,
    ) -> None:
        """Set properties."""

        self.image_fmt = image_fmt
        self.rel = rel
        self.dimensions = dimensions
        self.prefix = prefix

    @property
    def width(self) -> int:
        """Width from dimensions."""
        return self.dimensions[0]

    @property
    def height(self) -> int:
        """Height from dimensions."""
        return self.dimensions[1]

    def __repr__(self) -> str:
        """Representation of instance."""
        attr_names = (a for a in self.__dir__() if not a.startswith("_"))
        attrs = (f"{a}={getattr(self, a)}" for a in attr_names)
        return f"{self.__class__.__name__}({', '.join(attrs)})"

    def __str__(self) -> str:
        """Represent instance as string."""
        return self._get_filename_parts()

    def dict(self) -> FaviconDict:
        """Represent instance as dict."""
        return {
            "image_format": self.image_fmt,
            "dimensions": self.dimensions,
            "prefix": self.prefix,
            "rel": self.rel,
        }

    def json(self) -> str:
        """Represent instance as JSON string."""
        return _json.dumps(self.dict())

    def _get_filename_parts(self) -> str:
        """Don't add dimensions to favicon.ico."""
        parts: Tuple[str, ...] = (self.prefix,)

        if self.image_fmt == "ico":
            parts += (
                ".",
                self.image_fmt,
            )
        else:
            parts += (
                "-",
                "x".join(str(d) for d in self.dimensions),
                ".",
                self.image_fmt,
            )
        return "".join(parts)



import datetime
import re
import tomllib
from pathlib import Path
from typing import Any, Callable, TypeVar

from pydantic import TypeAdapter
from pydantic.json_schema import GenerateJsonSchema, JsonSchemaValue
from pydantic_core import CoreSchema
from pydantic_core.core_schema import NoneSchema, NullableSchema, StringSchema
from typing_extensions import TypedDict, override

from .assertions import cast_or_raise

# TODO: there's (figuratively) literally no comments in this file

TOMLValue = (
    str
    | int
    | float
    | bool
    | datetime.datetime
    | datetime.date
    | datetime.time
    | list["TOMLValue"]
    | dict[str, "TOMLValue"]
    | None
)

TOMLDict = dict[str, "TOMLValue"]


def fill_placeholders(data: TOMLDict):
    """Replaces hexdoc-style string placeholders in-place in a parsed TOML file."""
    _fill_placeholders(data, [data], set())


def _expand_placeholder(
    data: TOMLDict,
    stack: list[TOMLDict],
    expanded: set[tuple[int, str | int]],
    placeholder: str,
) -> str:
    tmp_stack: list[TOMLDict] = stack[:]

    # leading $ references the root table
    if placeholder[0] == "$":  # the regex *should* ensure placeholder is never empty
        placeholder = placeholder.lstrip("$.")
        tmp_stack = [tmp_stack[0]]

    key = "UNBOUND"
    keys = placeholder.split(".")
    for i, key in enumerate(keys):
        if n := key.count("^"):
            tmp_stack = tmp_stack[:-n]
            key = key.replace("^", "")
        if key and i < len(keys) - 1:
            # TODO: does this work?
            new = cast_or_raise(tmp_stack[-1][key], TOMLDict)
            tmp_stack.append(new)

    table = tmp_stack[-1]
    if (id(table), key) not in expanded:
        _handle_child(data, tmp_stack, expanded, key, table[key], table.__setitem__)

    value = cast_or_raise(table[key], str)
    return value


_T_key = TypeVar("_T_key", str, int)

_PLACEHOLDER_RE = re.compile(r"\{(.+?)\}")


def _handle_child(
    data: TOMLDict,
    stack: list[TOMLDict],
    expanded: set[tuple[int, str | int]],
    key: _T_key,
    value: TOMLValue,
    update: Callable[[_T_key, TOMLValue], None],
):
    # wait no that sounds wrong-
    match value:
        case str():
            # fill the string's placeholders
            for match in reversed(list(_PLACEHOLDER_RE.finditer(value))):
                try:
                    v = _expand_placeholder(
                        data,
                        stack,
                        expanded,
                        match[1],
                    )
                    value = value[: match.start()] + v + value[match.end() :]
                except Exception as e:
                    e.add_note(f"{match[0]} @ {value} @ {stack[-1]}")
                    raise
            expanded.add((id(stack[-1]), key))
            update(key, value)

        case {"!Raw": raw} if len(value) == 1:
            # interpolaten't
            expanded.add((id(stack[-1]), key))
            update(key, raw)

        case {"!None": _}:
            update(key, None)

        case list():
            # handle each item in the list without adding the list to the stack
            for i, item in enumerate(value):
                _handle_child(data, stack, expanded, i, item, value.__setitem__)

        case dict():
            # recursion!
            _fill_placeholders(data, stack + [value], expanded)

        case _:
            pass


def _fill_placeholders(
    data: TOMLDict,
    stack: list[TOMLDict],
    expanded: set[tuple[int, str | int]],
):
    table = stack[-1]
    for key, child in table.items():
        _handle_child(data, stack, expanded, key, child, table.__setitem__)


def load_toml_with_placeholders(path: Path) -> TOMLDict:
    data = tomllib.loads(path.read_text("utf-8"))
    fill_placeholders(data)
    return data


IntrinsicRaw = TypedDict("IntrinsicRaw", {"!Raw": Any})
raw_core_schema = TypeAdapter(IntrinsicRaw).core_schema

IntrinsicNone = TypedDict("IntrinsicNone", {"!None": Any})
none_core_schema = TypeAdapter(IntrinsicNone).core_schema


class GenerateJsonSchemaTOML(GenerateJsonSchema):
    _is_recursing = False

    @override
    def str_schema(self, schema: StringSchema) -> JsonSchemaValue:
        string_schema = super().str_schema(schema)
        if self._is_recursing:
            return string_schema

        return self.get_flattened_anyof(
            [
                string_schema,
                self._generate_inner_recursive(raw_core_schema),
            ]
        )

    @override
    def none_schema(self, schema: NoneSchema) -> JsonSchemaValue:
        return self.get_flattened_anyof(
            [
                super().none_schema(schema),
                self.generate_inner(none_core_schema),
            ]
        )

    @override
    def nullable_schema(self, schema: NullableSchema) -> JsonSchemaValue:
        null_schema = self.none_schema(NoneSchema(type="none"))
        inner_json_schema = self.generate_inner(schema["schema"])
        if inner_json_schema == null_schema:
            return null_schema
        return self.get_flattened_anyof([inner_json_schema, null_schema])

    def _generate_inner_recursive(self, core_schema: CoreSchema):
        # FIXME: hack
        self._is_recursing = True
        result = self.generate_inner(core_schema)
        self._is_recursing = False
        return result



from hexdoc.core import ResourceLocation
from hexdoc.minecraft import LocalizedStr
from hexdoc.minecraft.assets import ItemWithTexture, NamedTexture
from hexdoc.model import HexdocModel
from pydantic import Field

from .condition import Condition


def _gui_texture(name: str):
    return ResourceLocation("modonomicon", f"textures/gui/{name}.png")


class ParallaxLayer(HexdocModel):
    """https://klikli-dev.github.io/modonomicon/docs/basics/structure/categories#background_parallax_layers-json-array-of-json-objects-optional"""

    background: ResourceLocation
    speed: float
    vanish_zoom: float = 1


class Category(HexdocModel):
    """https://klikli-dev.github.io/modonomicon/docs/basics/structure/categories"""

    name: LocalizedStr
    icon: ItemWithTexture | NamedTexture

    sort_number: int = -1
    condition: Condition | None = None
    background: ResourceLocation = _gui_texture("dark_slate_seamless")
    background_parallax_layers: list[ParallaxLayer] = Field(default_factory=list)
    background_height: int = 512
    background_width: int = 512
    entry_textures: ResourceLocation = _gui_texture("entry_textures")
    show_category_button: bool = True



# 16x16 hashtag icon for tags
from yarl import URL

TAG_TEXTURE_URL = URL(
    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAACXBIWXMAAC4jAAAuIwF4pT92AAAANUlEQVQ4y2NgGJRAXV39v7q6+n9cfGTARKllFBvAiOxMUjTevHmTkSouGPhAHA0DWnmBrgAANLIZgSXEQxIAAAAASUVORK5CYII="
)

# purple and black square
MISSING_TEXTURE_URL = URL(
    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAIAAACQkWg2AAAACXBIWXMAAC4jAAAuIwF4pT92AAAAJElEQVQoz2NkwAF+MPzAKs7EQCIY1UAMYMQV3hwMHKOhRD8NAPogBA/DVsDEAAAAAElFTkSuQmCC"
)

NUM_GASLIGHTING_TEXTURES = 4
